/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 */

#include <config.h>

#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>

#include "nls.h"
#include "im-ja.h"
#include "conf.h"
#include "error.h"

#define GETWIDGET(s) glade_xml_get_widget(gladexml, s)
#define ENUMSIZE(s) sizeof(s)/sizeof(gchar *) 

static gchar *key_states[] = { "SHFT",
												"LOCK", 
												"CTRL",
												"MOD1",
												"MOD2", 
												"MOD3", 
												"MOD4", 
												"MOD5", 
												"BUT1",
												"BUT2",
												"BUT3",
												"BUT4",
												"BUT5",
												"BUT6"
};

static gchar *hotkey_labels[] = {
  "Start the im-ja configurator",
  "Switch to next input mode",
  "Switch to previous input mode",
  "Set raw input mode",
  "Toggle hiragana/katakana mode",
  "Toggle full/half-width katakana mode",
  "Toggle half/full-width katakana mode",
  "Toggle zenkaku/raw mode",
  "Toggle canna/raw mode",
  "Toggle wnn input mode",
  "Toggle kanji/raw mode",
  "Toggle kanjipad/raw input mode",
  "Unconvert current bunsetsu",
  "Convert current bunsetsu",
  "Get previous candidate",
  "Get next candidate",
  "Show all candidates",
  "Select previous bunsetsu",
  "Select next bunsetsu",
  "Expand current bunsetsu",
  "Shrink current bunsetsu"
};


static GtkWidget *window;
static GtkWidget *popup_window = NULL;
//static GtkWidget *hotkey_entry[HOTKEYS_TOTAL];
static im_ja_cfg cfg;
static GtkTreeModel *model;
static GtkTreeIter iter;
static GladeXML *gladexml;

extern gchar *hotkey_gconf_keys[];
extern GConfEnumStringPair status_win_labels[];

enum {
	HOTKEY_DESC,
	HOTKEY_KEYVAL,
	HOTKEY_IDX,
	NUM_COLS
};

enum {
	COLOR_NORMAL_FG,
	COLOR_NORMAL_BG,
	COLOR_HIGHL_FG,
	COLOR_HIGHL_BG
};


static gchar *get_keyname(guint state, guint keyval) {
  int i;
  gchar *keyname = g_new0(gchar, 140);

  //IM_JA_DEBUG("state: %d, keyval: %d, ", state, keyval);
  for (i = 0; i <= 13; i++) {
    if (state & (1 << i)) {
      //IM_JA_DEBUG("keyname: %s ", key_states[i]);
      g_strlcat(keyname, key_states[i], 140);
      g_strlcat(keyname, " ", 140);
    }
  }
  //IM_JA_DEBUG("%s\n", gdk_keyval_name(keyval));
		if ((keyval == 0) || (keyval == GDK_VoidSymbol)) {
			g_strlcat(keyname, _("Undefined"), 140);
		}
		else g_strlcat(keyname, gdk_keyval_name(keyval), 140);
  if (strlen(keyname) == 0) g_strlcat(keyname, "???", 140);
  return keyname;
}

static gboolean get_key(GtkWidget *widget, GdkEventKey *event, gpointer data) {
  guint state;
  guint keyval;
  gchar *keystr;
  int idx;
  
  if (event != NULL) {
    if ((event)->type == GDK_KEY_RELEASE) {
      gdk_keyboard_ungrab(GDK_CURRENT_TIME);
      /* We don't want Num lock, caps lock and scroll lock */
      keyval = event->keyval;
      state = event->state;
      state &= ~GDK_LOCK_MASK; 
      state &= ~GDK_MOD2_MASK;
      state &= ~GDK_MOD5_MASK;
      gtk_tree_model_get(model, &iter, HOTKEY_IDX, &idx, -1);
      IM_JA_DEBUG("selected: %d\n", idx);
      cfg.hotkey_values[idx] = keyval;
      cfg.hotkey_states[idx] = state;
      keystr = get_keyname(state, keyval);
      IM_JA_DEBUG("GOT KEY: keyval = %d, keyval_name = %s, state = %d str = \"%s\" \n", keyval, gdk_keyval_name(keyval), state, keystr);
      if (keyval == GDK_Return) return FALSE;
      gtk_list_store_set(GTK_LIST_STORE(model), &iter, HOTKEY_KEYVAL, keystr, -1);
      g_free(keystr);
    }
  }
  gtk_widget_hide(popup_window);
  gtk_widget_destroy(popup_window);
  return TRUE;
}


static void do_keygrab() {
  GtkWidget *button_cancel;

  popup_window = gtk_message_dialog_new(GTK_WINDOW(window), GTK_DIALOG_DESTROY_WITH_PARENT, 
																				GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, _("waiting for keypress..."));
  gtk_window_set_modal(GTK_WINDOW(popup_window), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(popup_window), GTK_WINDOW(window));
  button_cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_dialog_add_action_widget(GTK_DIALOG(popup_window), button_cancel, 1);

  g_signal_connect(G_OBJECT(popup_window), "key-release-event", G_CALLBACK(get_key), NULL);
  g_signal_connect_swapped(GTK_OBJECT(popup_window), "response",
                           G_CALLBACK(gtk_widget_destroy),
                           GTK_OBJECT(popup_window));
  GTK_WIDGET_UNSET_FLAGS(button_cancel, GTK_CAN_FOCUS);

  gtk_widget_show(button_cancel);
  gtk_widget_show(popup_window);
}

static void activate_cb(GtkTreeView *treeview, GtkTreePath *treepath, GtkTreeViewColumn *column, gpointer data) {
  gtk_tree_model_get_iter(model, &iter, treepath);
  do_keygrab();
}

/*
	static void click_cb(GtkTreeView *treeview, gpointer data) {
  GtkTreeSelection *tree_selection;

  tree_selection = gtk_tree_view_get_selection(treeview);
  if (gtk_tree_selection_get_selected (tree_selection, &model, &iter) == FALSE) return;
  do_keygrab();
	}

	static gboolean select_cb(GtkTreeSelection *tree_selection, gpointer data) {
  IM_JA_DEBUG("selection in list.\n");
  if (gtk_tree_selection_get_selected(tree_selection, &model, &iter) == FALSE) return FALSE;
  if (gtk_tree_selection_iter_is_selected(tree_selection, &iter) == FALSE) return FALSE;
  IM_JA_DEBUG("rows selected: %d\n", gtk_tree_selection_count_selected_rows(tree_selection));
  do_keygrab();
  return TRUE;
	}
*/

static void dialog_response_cb(GtkDialog *dialog, gint response_id) {
	int i;
	gchar *tmpstr;

  if (response_id == GTK_RESPONSE_APPLY) {
		cfg.status_win_coords[0] = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_status_xpos")));
		cfg.status_win_coords[1] = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_status_ypos")));

		cfg.kanjipad_size[0] = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_kanjipad_width")));
		cfg.kanjipad_size[1] = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_kanjipad_height")));

		for (i = 0; i < IM_JA_INPUT_METHODS_TOTAL; i++) {
			tmpstr = g_strdup_printf("entry_status_label_%d", i);
			cfg.status_win_labels[i] = g_strdup(gtk_entry_get_text(GTK_ENTRY(GETWIDGET(tmpstr))));
			g_free(tmpstr);
		}

		cfg.wnn_address = g_strdup(gtk_entry_get_text(GTK_ENTRY(GETWIDGET("entry_wnn_address"))));
		//		cfg.wnn_port = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_wnn_port")));
		cfg.wnnenvrc = g_strdup(gtk_entry_get_text(GTK_ENTRY(GETWIDGET("entry_wnnenvrc"))));

		//		cfg.canna_address = g_strdup(gtk_entry_get_text(GTK_ENTRY(GETWIDGET("entry_canna_address"))));
		//		cfg.canna_port = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_canna_port")));

    IM_JA_DEBUG("save config.\n");
    im_ja_save_conf(&cfg);
  }
  else {
		gtk_main_quit();
  }
}

static void set_preview_style() {
	gtk_widget_modify_text(GETWIDGET("preview_n_entry"), GTK_STATE_NORMAL, &cfg.preedit_colors[0]);
	gtk_widget_modify_base(GETWIDGET("preview_n_entry"), GTK_STATE_NORMAL, &cfg.preedit_colors[1]);
	gtk_widget_modify_text(GETWIDGET("preview_hl_entry"), GTK_STATE_NORMAL, &cfg.preedit_colors[2]);
	gtk_widget_modify_base(GETWIDGET("preview_hl_entry"), GTK_STATE_NORMAL, &cfg.preedit_colors[3]);
}

static void change_color_cb(GtkWidget *button, int which) {
  GtkWidget *dialog;
  GtkColorSelection *colorsel;
  gint response;
  GdkColor *cfgcolor = &cfg.preedit_colors[which];

	dialog = gtk_color_selection_dialog_new(_("Color selection"));

  gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(window));
  
  colorsel = GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(dialog)->colorsel);

	gtk_color_selection_set_previous_color(colorsel, cfgcolor);
	gtk_color_selection_set_current_color(colorsel, cfgcolor);
  
  gtk_color_selection_set_has_palette(colorsel, TRUE);
  
  response = gtk_dialog_run(GTK_DIALOG(dialog));

  if (response == GTK_RESPONSE_OK) {
    IM_JA_DEBUG("Color selection for %d\n", which);
    gtk_color_selection_get_current_color(colorsel, cfgcolor);
		set_preview_style();
  }
  gtk_widget_destroy(dialog);
}

static void radiobutton_n_cb(GtkWidget *button) {
	gboolean iscustom  = !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_n_default")));
	
	gtk_widget_set_sensitive(GETWIDGET("preview_n_entry"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("preview_n_label"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_n_b_button"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_n_f_button"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_n_b_label"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_n_f_label"), iscustom);

	cfg.custom_preedit_n = iscustom;
}

static void radiobutton_hl_cb(GtkWidget *button) {
	gboolean iscustom  = !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_hl_default")));
	
	gtk_widget_set_sensitive(GETWIDGET("preview_hl_entry"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("preview_hl_label"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_hl_b_button"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_hl_f_button"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_hl_b_label"), iscustom);
	gtk_widget_set_sensitive(GETWIDGET("colsel_hl_f_label"), iscustom);

	cfg.custom_preedit_hl = iscustom;
}

static void radiobutton_status_win_pos_cb(GtkWidget *button) {
	gboolean fixed = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_fixed_coord")));
	
	gtk_widget_set_sensitive(GETWIDGET("label_status_xpos"), fixed);
	gtk_widget_set_sensitive(GETWIDGET("label_status_ypos"), fixed);
	gtk_widget_set_sensitive(GETWIDGET("spinbutton_status_xpos"), fixed);
	gtk_widget_set_sensitive(GETWIDGET("spinbutton_status_ypos"), fixed);
													 
	
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_below_cursor"))) == TRUE) {
		cfg.status_win_position = STATUS_BELOW_CURSOR;
		return;
	}
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_window_bottom_l"))) == TRUE) {
		cfg.status_win_position = STATUS_WINDOW_BOTTOM_LEFT;	
		return;
	}
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_fixed_coord"))) == TRUE) {
		cfg.status_win_position = STATUS_FIXED_COORD;
		return;
	}
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_status_disabled"))) == TRUE) {
		cfg.status_win_position = STATUS_DISABLED;
	}
}

static void radiobutton_conv_engine_cb(GtkWidget *button) {
	gboolean wnn_default = FALSE, canna_default = FALSE;


	wnn_default = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_wnn_default")));
	canna_default = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_canna_default")));

#ifdef DISABLE_WNN
	wnn_default = FALSE;
#endif
#ifdef DISABLE_CANNA
	canna_default = FALSE;
#endif

	gtk_widget_set_sensitive(GETWIDGET("label_wnn_address"), wnn_default);
	gtk_widget_set_sensitive(GETWIDGET("label_wnnenvrc"), wnn_default);
	//	gtk_widget_set_sensitive(GETWIDGET("label_wnn_port"), wnn_default);
	gtk_widget_set_sensitive(GETWIDGET("entry_wnn_address"), wnn_default);
	//	gtk_widget_set_sensitive(GETWIDGET("spinbutton_wnn_port"), wnn_default);
	gtk_widget_set_sensitive(GETWIDGET("entry_wnnenvrc"), wnn_default);
	gtk_widget_set_sensitive(GETWIDGET("label_wnn_text"), wnn_default);
	/*
	gtk_widget_set_sensitive(GETWIDGET("label_canna_address"), canna_default);
	gtk_widget_set_sensitive(GETWIDGET("label_canna_port"), canna_default);
	gtk_widget_set_sensitive(GETWIDGET("entry_canna_address"), canna_default);
	gtk_widget_set_sensitive(GETWIDGET("spinbutton_canna_port"), canna_default);
	gtk_widget_set_sensitive(GETWIDGET("label_canna_text"), canna_default);
	*/
	if (wnn_default == TRUE) cfg.default_conv_engine = KANJI_INPUT_WNN;
	if (canna_default == TRUE) cfg.default_conv_engine = KANJI_INPUT_CANNA;

}

static void checkbutton_enable_kanjipad_cb(GtkWidget *button) {
	gboolean enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("checkbutton_enable_kanjipad")));

	if (((cfg.kanjipad_custom_size == TRUE) && (enabled == TRUE)) || (enabled == FALSE)) {
		gtk_widget_set_sensitive(GETWIDGET("spinbutton_kanjipad_width"), enabled);
		gtk_widget_set_sensitive(GETWIDGET("spinbutton_kanjipad_height"), enabled);
		gtk_widget_set_sensitive(GETWIDGET("label_kanjipad_width"), enabled);
		gtk_widget_set_sensitive(GETWIDGET("label_kanjipad_height"), enabled);
	}
	gtk_widget_set_sensitive(GETWIDGET("checkbutton_kanjipad_size"), enabled);
	
	cfg.kanjipad_enabled = enabled;

}

static void checkbutton_kanjipad_size_cb(GtkWidget *button) {
	gboolean enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(GETWIDGET("checkbutton_kanjipad_size")));

	gtk_widget_set_sensitive(GETWIDGET("spinbutton_kanjipad_width"), enabled);
	gtk_widget_set_sensitive(GETWIDGET("spinbutton_kanjipad_height"), enabled);
	gtk_widget_set_sensitive(GETWIDGET("label_kanjipad_width"), enabled);
	gtk_widget_set_sensitive(GETWIDGET("label_kanjipad_height"), enabled);

	cfg.kanjipad_custom_size = enabled;
}

static void startup_menu_cb(gint input_method) {
	IM_JA_DEBUG("startup selected: %d.\n", input_method);
	cfg.startup_input_method = input_method;
}

static GtkWidget *create_dialog() {
  GtkWidget *treeview;
	GtkWidget *menu_item;
	GtkWidget *menu;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;
	gint i;
  gchar *tmpstr;

  gladexml = glade_xml_new(KP_LIBDIR"/im-ja-conf.glade", "notebook", NULL);
  
	//	glade_xml_signal_autoconnect(gladexml);

  /* Set up the table of hotkey assignments */
	treeview = GETWIDGET("treeview");

  model = GTK_TREE_MODEL(gtk_list_store_new(NUM_COLS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT));
  gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
	gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), model);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Action"), renderer, "text", HOTKEY_DESC, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Key"), renderer, "text", HOTKEY_KEYVAL, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
  g_signal_connect(G_OBJECT(treeview), "row_activated", G_CALLBACK(activate_cb), NULL);

  for (i = 0; i < ENUMSIZE(hotkey_labels); i++) {
		/* Don't show canna and wnn keys in configurator */
		if ((i != 8) && (i != 9)) { // FIXME: use GConfEnumStringPair constant
			gtk_list_store_append(GTK_LIST_STORE(model), &iter);
			tmpstr = get_keyname(cfg.hotkey_states[i], cfg.hotkey_values[i]);
			gtk_list_store_set(GTK_LIST_STORE(model), &iter, 
												 HOTKEY_DESC, hotkey_labels[i], 
												 HOTKEY_KEYVAL, tmpstr,
												 HOTKEY_IDX, i,
												 -1);
			IM_JA_DEBUG("adding %d: %s\n", i, hotkey_labels[i]);
			g_free(tmpstr);
		}
  }

  /* Set up color pickers */
  g_signal_connect(GTK_OBJECT(GETWIDGET("colsel_n_f_button")), "clicked",
									 G_CALLBACK(change_color_cb),
									 (gpointer) COLOR_NORMAL_FG);
  g_signal_connect(GTK_OBJECT(GETWIDGET("colsel_n_b_button")), "clicked",
									 G_CALLBACK(change_color_cb),
									 (gpointer) COLOR_NORMAL_BG);
  g_signal_connect(GTK_OBJECT(GETWIDGET("colsel_hl_f_button")), "clicked",
									 G_CALLBACK(change_color_cb),
									 (gpointer) COLOR_HIGHL_FG);
  g_signal_connect(GTK_OBJECT(GETWIDGET("colsel_hl_b_button")), "clicked",
									 G_CALLBACK(change_color_cb),
									 (gpointer) COLOR_HIGHL_BG);

	/* Set up custom color radio buttons */
	if (cfg.custom_preedit_n == TRUE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_n_custom")), TRUE);	
	}
	else gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_n_default")), TRUE);	
	
	if (cfg.custom_preedit_hl == TRUE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_hl_custom")), TRUE);	
	}
	else gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_hl_default")), TRUE);	
	
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_hl_default")), "toggled",
									 G_CALLBACK(radiobutton_hl_cb), NULL);
  g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_n_default")), "toggled",
									 G_CALLBACK(radiobutton_n_cb), NULL);
	
	radiobutton_n_cb(NULL);
	radiobutton_hl_cb(NULL);
	
	/* Set up status window position radio buttons */
	switch (cfg.status_win_position) {
	case STATUS_BELOW_CURSOR:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_below_cursor")), TRUE);	
		break;
	case STATUS_WINDOW_BOTTOM_LEFT:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_window_bottom_l")), TRUE);	
		break;
	case STATUS_FIXED_COORD:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_fixed_coord")), TRUE);	
		break;
	case STATUS_DISABLED:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_status_disabled")), TRUE);	
		break;
	}

	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_below_cursor")), "toggled",
									 G_CALLBACK(radiobutton_status_win_pos_cb), NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_window_bottom_l")), "toggled",
									 G_CALLBACK(radiobutton_status_win_pos_cb), NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_fixed_coord")), "toggled",
									 G_CALLBACK(radiobutton_status_win_pos_cb), NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_status_disabled")), "toggled",
									 G_CALLBACK(radiobutton_status_win_pos_cb), NULL);
	radiobutton_status_win_pos_cb(NULL);	

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_status_xpos")), cfg.status_win_coords[0]);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_status_ypos")), cfg.status_win_coords[1]);

	for (i = 0; i < IM_JA_INPUT_METHODS_TOTAL; i++) {
		tmpstr = g_strdup_printf("entry_status_label_%d", i);
		if (cfg.status_win_labels[i] != NULL) gtk_entry_set_text(GTK_ENTRY(GETWIDGET(tmpstr)), cfg.status_win_labels[i]);
		g_free(tmpstr);
	}

	switch (cfg.default_conv_engine) {
	case KANJI_INPUT_WNN:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_wnn_default")), TRUE);	
		break;
	case KANJI_INPUT_CANNA:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("radiobutton_canna_default")), TRUE);	
		break;
	}
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_wnn_default")), "toggled",
									 G_CALLBACK(radiobutton_conv_engine_cb), NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("radiobutton_canna_default")), "toggled",
									 G_CALLBACK(radiobutton_conv_engine_cb), NULL);
	radiobutton_conv_engine_cb(NULL);
#ifdef DISABLE_WNN
	gtk_widget_set_sensitive(GETWIDGET("radiobutton_wnn_default"), FALSE);
#endif
#ifdef DISABLE_CANNA
	gtk_widget_set_sensitive(GETWIDGET("radiobutton_canna_default"), FALSE);
#endif

	//	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_wnn_port")), cfg.wnn_port);
	//	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_canna_port")), cfg.canna_port);
	if (cfg.wnn_address != NULL) gtk_entry_set_text(GTK_ENTRY(GETWIDGET("entry_wnn_address")), cfg.wnn_address);
	if (cfg.wnnenvrc != NULL) gtk_entry_set_text(GTK_ENTRY(GETWIDGET("entry_wnnenvrc")), cfg.wnnenvrc);
	//	if (cfg.canna_address != NULL) gtk_entry_set_text(GTK_ENTRY(GETWIDGET("entry_canna_address")), cfg.canna_address);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("checkbutton_enable_kanjipad")), cfg.kanjipad_enabled);	
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GETWIDGET("checkbutton_kanjipad_size")), cfg.kanjipad_custom_size);	
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_kanjipad_width")), cfg.kanjipad_size[0]);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(GETWIDGET("spinbutton_kanjipad_height")), cfg.kanjipad_size[1]);
	checkbutton_enable_kanjipad_cb(NULL);
	checkbutton_kanjipad_size_cb(NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("checkbutton_enable_kanjipad")), "toggled",
									 G_CALLBACK(checkbutton_enable_kanjipad_cb), NULL);
	g_signal_connect(GTK_OBJECT(GETWIDGET("checkbutton_kanjipad_size")), "toggled",
									 G_CALLBACK(checkbutton_kanjipad_size_cb), NULL);

	
  menu = gtk_menu_new();

	for (i = 0; i < IM_JA_INPUT_METHODS_TOTAL; i++) {
		menu_item = gtk_menu_item_new_with_label(gconf_enum_to_string(status_win_labels, i));
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
		g_signal_connect_swapped(GTK_OBJECT(menu_item), "activate", 
														 GTK_SIGNAL_FUNC(startup_menu_cb), (gpointer) i);
    gtk_widget_show(menu_item);
	}
  gtk_option_menu_set_menu(GTK_OPTION_MENU(GETWIDGET("optionmenu_startup_mode")), menu);
	gtk_option_menu_set_history(GTK_OPTION_MENU(GETWIDGET("optionmenu_startup_mode")), cfg.startup_input_method);

	return GETWIDGET("notebook");
}



int main(int argc, char *argv[]) {
  GtkWidget *dialog;
  GtkWidget *dialog_win;

#ifdef ENABLE_NLS
  bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
  textdomain(GETTEXT_PACKAGE);
#endif

  gtk_init(&argc, &argv);
    
  if (im_ja_init_conf_handler() == FALSE) {
    im_ja_print_error(_("GConf initialization failed!"));
  }
  if (im_ja_load_conf(&cfg) == FALSE) {
    im_ja_print_error(_("Couldn't load settings!"));
  }

	dialog = create_dialog();
	set_preview_style();
  
  dialog_win = gtk_dialog_new_with_buttons(_("im-ja preferences"), NULL, 0,
																					 GTK_STOCK_APPLY, GTK_RESPONSE_APPLY,
																					 GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
																					 NULL);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog_win),
																	GTK_RESPONSE_CLOSE);
  
  g_signal_connect(G_OBJECT(dialog_win), "response",
									 G_CALLBACK(dialog_response_cb), NULL);
  
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog_win)->vbox), 
										 dialog, TRUE, TRUE, 10);
	gtk_widget_show(dialog_win);

  gtk_main();
  
  return 0;
}
