/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 * Based on code by:
 *          Owen Taylor <otaylor@redhat.com>
 *          Yusuke Tabata <tee@kuis.kyoto-u.ac.jp>
 *          Yukihiro Nakai <ynakai@redhat.com>
 *          
 */
#include <config.h>
#include <configure.h>

#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>

GType type_ja = 0;
#define TYPE_JA_DEFINED

#include "im-ja.h"
#ifndef DISABLE_CANNA
#include "canna_rk.h"
#endif
#include "common.h"
#include "error.h"
#include "conf.h"
#include "nls.h"
#include "romakana.h"

#ifndef DISABLE_WNN
#include "wnn.h"
#endif

static const GtkIMContextInfo im_ja_info = { 
	"im-ja",			/* ID */
	"Japanese",							/* Human readable name */
	"gtk+",			/* Translation domain */
	"",							/* Dir for bindtextdomain (not strictly needed for "gtk+") */
	"ja"				/* Languages for which this module is the default */
};

static const GtkIMContextInfo *info_list[] = {
	&im_ja_info
};


typedef struct _IMContextJAClass {
	GtkIMContextClass parent_class;
} IMContextJAClass;

enum {
	STRING_COLUMN,
	N_COLUMNS
};

im_ja_cfg cfg; 
//static gboolean wnn_input_usable = TRUE;
//static gboolean canna_input_usable = TRUE;

/* * * * * * * * * * * */

/* GtkIMContext API */
static void im_ja_class_init(GtkIMContextClass *class);
static void im_ja_class_finalize(GObject *obj);
static void im_ja_init(GtkIMContext *im_context);
static void im_ja_finalize(GtkIMContext *context);
static gboolean im_ja_filter_keypress(GtkIMContext *context, GdkEventKey *key);
static void im_ja_get_preedit_string(GtkIMContext *context, gchar **str, PangoAttrList **attrs, gint *cursor_pos);
static void im_ja_cursor_location_changed(GtkIMContext *context, GdkRectangle *area);
static void im_ja_set_client_window(GtkIMContext* context, GdkWindow *win);
static void im_ja_lost_focus(GtkIMContext *context);
static void im_ja_got_focus(GtkIMContext *context);
static void im_ja_set_use_preedit(GtkIMContext *context, gboolean use_preedit);
static void im_ja_on_reset(IMContextJA *context);
void im_ja_update_status_window_location(IMContextJA *context);
void im_ja_set_kanjipad_location(IMContextJA *context);
void disable_status_window_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA *context);
void disable_kanjipad_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA *context);
gboolean im_ja_set_kanji_input(IMContextJA *context, gint conv_engine);

GtkWidget *create_kanjipad_win(IMContextJA *context, gint width, gint height);

/* * * * * * * * * * * */


static void im_ja_register_type (GTypeModule *module) {
	static const GTypeInfo object_info = {
		sizeof (IMContextJAClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) im_ja_finalize,
		(GClassInitFunc) im_ja_class_init,
		(GClassFinalizeFunc) im_ja_class_finalize,
		NULL,						/* class_data */
		sizeof (IMContextJA),
		0,
		(GtkObjectInitFunc) im_ja_init,
	};
	
	type_ja =	 g_type_module_register_type(module, GTK_TYPE_IM_CONTEXT,
																				 "IMContextJA", &object_info, 0);
}

/* This is called only once for the application */
static void im_ja_class_init (GtkIMContextClass *class) {
	GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(class);
	GObjectClass *object_class = G_OBJECT_CLASS(class);
	IM_JA_DEBUG("im_ja_class_init\n");
	im_context_class->set_client_window = im_ja_set_client_window;
	im_context_class->filter_keypress = im_ja_filter_keypress;
	im_context_class->get_preedit_string = im_ja_get_preedit_string;
	im_context_class->set_cursor_location = im_ja_cursor_location_changed;
	im_context_class->focus_in = (void *)im_ja_got_focus;
	im_context_class->focus_out = (void *)im_ja_lost_focus;
	im_context_class->reset = (void*)im_ja_on_reset;
	im_context_class->set_use_preedit = (void *)im_ja_set_use_preedit;
	object_class->finalize = im_ja_class_finalize;

	if (im_ja_init_conf_handler() == FALSE) {
		im_ja_print_error(_("GConf initialization failed!"));
	}
	if (im_ja_load_conf(&cfg) == FALSE) {
		im_ja_print_error(_("Couldn't load settings!"));
	}
}

static void im_ja_class_finalize(GObject *obj) {
	IM_JA_DEBUG("FIXME: im-ja class finalize\n");
}

static void im_ja_finalize(GtkIMContext *context) {
	//	IMContextJA *cn = IM_CONTEXT_JA(context);
	IM_JA_DEBUG("FIXME: im-ja finalize\n");
}

/* Called for every input widget instance */
static void im_ja_init(GtkIMContext *context) {
	IMContextJA *cn = IM_CONTEXT_JA(context);

	IM_JA_DEBUG("im_ja_init\n");
	cn->kanjipad = NULL;
	cn->preedit_buf = g_new0(guchar, BUFFERSIZE);
	cn->preedit_insert = g_new0(guchar, BUFFERSIZE);
	cn->kakutei_buf = g_new0(guchar, BUFFERSIZE);
	//cn->canna_diclist = g_new0(guchar, BUFFERSIZE);
	cn->client_window = NULL;
	cn->status_window = NULL;
	cn->preedit_reverse_start = 0;
	cn->preedit_reverse_end = 0;
	cn->cursor_ndx = -1;
	cn->show_first = FALSE;
	cn->on_reset = NULL;
	im_ja_set_input_method(cn, cfg.startup_input_method);
}

void im_module_init(GTypeModule *module) {
	im_ja_register_type(module);
}

void im_module_exit(void) {
}

void im_module_list(const GtkIMContextInfo ***contexts, int *n_contexts) {
	*contexts = info_list;
	*n_contexts = G_N_ELEMENTS(info_list);
}

GtkIMContext *im_module_create (const gchar *context_id) {
	if (strcmp(context_id, "im-ja") == 0) {
		return GTK_IM_CONTEXT(g_object_new(type_ja, NULL));
	}
	else {
		return NULL;
	}
}

static void im_ja_set_use_preedit(GtkIMContext *context, gboolean use_preedit) {
	IM_JA_DEBUG("im_ja_set_use_preedit: %d\n", use_preedit);
}

gboolean im_ja_filter_keypress(GtkIMContext *context, GdkEventKey *key) {
	gboolean retval = FALSE;
	IMContextJA *cn = IM_CONTEXT_JA(context);

	if (key->type == GDK_KEY_RELEASE) return FALSE;

	/*
		IM_JA_DEBUG("GOT KEYEVENT. keyname: %s, keyval : %d,	state: %d\n", 
		gdk_keyval_name(key->keyval), key->keyval, key->state);
	*/

	if (ishotkey(key, START_CONFIGURATOR, &cfg) == TRUE) {
		gchar *command[] = {BINDIR"/im-ja-conf"};
		GError *g_error = NULL;
		if (g_spawn_async(NULL, command, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
											NULL, NULL, NULL, &g_error) == FALSE) {
			im_ja_print_error(g_error->message); 
		}
		return TRUE;
	}
	if (ishotkey(key, NEXT_INPUT_MODE, &cfg) == TRUE) {
		im_ja_next_input_method(cn);
		return TRUE;
	}
	if (ishotkey(key, PREV_INPUT_MODE, &cfg) == TRUE) {
		im_ja_prev_input_method(cn);
		return TRUE;
	}
	if (ishotkey(key, RAW_INPUT_MODE, &cfg) == TRUE) {
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return TRUE;
	}
	if (ishotkey(key, HIRAGANA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already hiragana, then set to katakana */
		if (cn->input_method == IM_JA_HIRAGANA_INPUT) im_ja_set_input_method(cn, IM_JA_KATAKANA_INPUT);
		else im_ja_set_input_method(cn, IM_JA_HIRAGANA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, KATAKANA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already katakana, then set to half-widh katakana */
		if (cn->input_method == IM_JA_KATAKANA_INPUT) im_ja_set_input_method(cn, IM_JA_HALFKATA_INPUT);
		else im_ja_set_input_method(cn, IM_JA_KATAKANA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, HALFKATA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already half-katakana, then set to full katakana */
		if (cn->input_method == IM_JA_HALFKATA_INPUT) im_ja_set_input_method(cn, IM_JA_KATAKANA_INPUT);
		else im_ja_set_input_method(cn, IM_JA_HALFKATA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, ZENKAKU_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already zenkaku, then set to RAW */
		if (cn->input_method == IM_JA_ZENKAKU_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_ZENKAKU_INPUT);
		return TRUE;
	}
	if (ishotkey(key, KANJI_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already KANJI, then set to RAW */
		if (cn->input_method == IM_JA_KANJI_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_KANJI_INPUT);
		return TRUE;
	}
	if (ishotkey(key, CANNA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already CANNA, then set to RAW */
		if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_CANNA)) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_kanji_input(cn, KANJI_INPUT_CANNA);
		return TRUE;
	}
	if (ishotkey(key, WNN_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already WNN, then set to RAW */
		if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_WNN)) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_kanji_input(cn, KANJI_INPUT_WNN);
		return TRUE;
	}
	if (ishotkey(key, KANJIPAD_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already KANJIPAD, then set to RAW */
		if (cn->input_method == IM_JA_KANJIPAD_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_KANJIPAD_INPUT);
		return TRUE;
	}

	switch (cn->input_method) {
	case IM_JA_RAW_INPUT: //FIXME: Is there any way to do _real_ raw input instead of emulating it?
		// maybe gdk_display_put_event ??
	case IM_JA_KANJIPAD_INPUT:
		if (key->state & GDK_CONTROL_MASK) return FALSE; 
	 
		if (key->keyval >= GDK_exclam && key->keyval <= GDK_overline) { 
			gchar utf8strg[7];
			int utf8strg_end;
			utf8strg_end = g_unichar_to_utf8(gdk_keyval_to_unicode(key->keyval), utf8strg);
			utf8strg[utf8strg_end] = 0;
			g_signal_emit_by_name(cn, "commit", utf8strg);
			return TRUE;
		}
		if (key->keyval == GDK_space) {
			g_signal_emit_by_name(cn, "commit", " ");
			return TRUE;
		}
		return FALSE;
	case IM_JA_HIRAGANA_INPUT:
	case IM_JA_KATAKANA_INPUT:
	case IM_JA_HALFKATA_INPUT:
	case IM_JA_ZENKAKU_INPUT:
		retval = im_ja_kana_filter_keypress(cn, key);
		break;
	case IM_JA_KANJI_INPUT:
		if (cfg.default_conv_engine == KANJI_INPUT_CANNA) {
#ifndef DISABLE_CANNA
			retval = canna_rk_filter_keypress(cn, key);
#else
			IM_JA_DEBUG("This shouldn't happen. We are not supposed to be in canna mode if it's disabled\n");
#endif
		}
		else if (cfg.default_conv_engine == KANJI_INPUT_WNN) {
#ifndef DISABLE_WNN
			retval = im_ja_wnn_filter_keypress(cn, key);
#else
			IM_JA_DEBUG("This shouldn't happen. We are not supposed to be in wnn mode if it's disabled\n");
#endif
		break;
		}
	}

	/* ESC pressed: drop current pre-edit buffer if this keypress wasn't handled */
	if ((key->keyval == GDK_Escape) && (retval == FALSE)) {
		IM_JA_DEBUG("ESC: cancel preedit\n");
		IM_JA_DEBUG("preedit_insert: %d\n", strlen(cn->preedit_insert));
		IM_JA_DEBUG("preedit_buf: %d\n", strlen(cn->preedit_buf));
		if ((strlen(cn->preedit_buf) > 0) || (strlen(cn->preedit_insert) > 0)) { //Cancel preedit buffer
			if (cn->on_reset) (*cn->on_reset)(cn);
			im_ja_reset(cn);
		}
		else {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		}
		return TRUE;
	}

	/*
	 * This block was is perhaps ill advised, gtk should give us
	 * reset events whenever necessary
	 * 
	 //ignore shifts, locks, etc. they are not actual keypresses
	 if ((key->keyval >= GDK_Shift_L && key->keyval <= GDK_Hyper_R) ||
	 key->keyval == GDK_ISO_Group_Shift ||
	 key->keyval == GDK_Num_Lock ||
	 key->keyval == GDK_Scroll_Lock) {
	 IM_JA_DEBUG("Ignoring \"shift\" type key:0x%X\n", key->keyval);
	 return FALSE;
	 }

	 //lets be conservative: cancel preedit anytime we
	 //dont know what key was used
	 if ((strlen(cn->preedit_buf) > 0) && (retval == FALSE)) {
	 IM_JA_DEBUG("Not caught:0x%X, killing preeditbuffer\n", key->keyval);
	 //Cancel preedit buffer
	 if (cn->on_reset) (*cn->on_reset)(cn);
	 im_ja_reset(cn);
	 return TRUE;
	 }
	*/

	return retval;
}

void im_ja_set_preedit_string(IMContextJA *cn, gchar *str, gint reverse_start, gint reverse_end) {
	//	printf("set preedit str\n");
	if (str == NULL) memset(cn->preedit_buf, 0, BUFFERSIZE);
	else g_strlcpy(cn->preedit_buf, str, BUFFERSIZE);
	cn->preedit_reverse_start = reverse_start;
	cn->preedit_reverse_end = reverse_end;
}


void im_ja_get_preedit_string(GtkIMContext *ic, gchar **str, PangoAttrList **attrs, gint *cursor_pos) {
	IMContextJA *cn = IM_CONTEXT_JA(ic);

	if (attrs != NULL) {
		*attrs = pango_attr_list_new(); //Create new attrib list
	}

	if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
		im_ja_print_error("utf8_validate failed for preedit string: %s\n", cn->preedit_buf);
	}
	else *str = g_strdup(cn->preedit_buf);
	
	if (*str == NULL || strlen(*str) == 0) return;

	if (attrs != NULL) {
		/* Normal (and underlined) preedit string */
		PangoAttribute *backgr;
		PangoAttribute *foregr;
		PangoAttribute *underline;

		underline = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
		underline->start_index = 0;
		underline->end_index = strlen(*str);
		pango_attr_list_insert(*attrs, underline);
		
		if (cfg.custom_preedit_n == TRUE) {
			foregr = pango_attr_foreground_new(cfg.preedit_colors[0].red, cfg.preedit_colors[0].green, cfg.preedit_colors[0].blue);
			backgr = pango_attr_background_new(cfg.preedit_colors[1].red, cfg.preedit_colors[1].green, cfg.preedit_colors[1].blue);
			backgr->start_index = 0;
			foregr->start_index = 0;
			backgr->end_index = strlen(*str);
			foregr->end_index = strlen(*str);
			pango_attr_list_insert(*attrs, backgr);
			pango_attr_list_insert(*attrs, foregr);
		}

		/* Highlighted (==reverse) field */
		if (cfg.custom_preedit_hl == TRUE) {
			foregr = pango_attr_foreground_new(cfg.preedit_colors[2].red, cfg.preedit_colors[2].green, cfg.preedit_colors[2].blue);
			backgr = pango_attr_background_new(cfg.preedit_colors[3].red, cfg.preedit_colors[3].green, cfg.preedit_colors[3].blue);
		}
		else { /* Use the default */
			foregr = pango_attr_foreground_new(cfg.original_colors[0].red, cfg.original_colors[0].green, cfg.original_colors[0].blue);
			backgr = pango_attr_background_new(cfg.original_colors[1].red, cfg.original_colors[1].green, cfg.original_colors[1].blue);
		}
		backgr->start_index = cn->preedit_reverse_start;
		foregr->start_index = cn->preedit_reverse_start;
		backgr->end_index = cn->preedit_reverse_end;
		foregr->end_index = cn->preedit_reverse_end;

		pango_attr_list_insert(*attrs, backgr);
		pango_attr_list_insert(*attrs, foregr);
	}

	if (cursor_pos != NULL) {
		if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_WNN)) {
			*cursor_pos = cn->cursor_wnn;
		}
		else {
			if (cn->cursor_ndx == -1) {
				*cursor_pos = strlen(*str);
			}
			else {
				//convert byte index into character index
				*cursor_pos = g_utf8_strlen(*str, cn->cursor_ndx);
			}
		}
		IM_JA_DEBUG("im_ja_get_preedit_string: cursor char pos: %d\n", *cursor_pos);
	}

}


static void im_ja_set_client_window(GtkIMContext *context, GdkWindow *win) {
	IMContextJA* cn = IM_CONTEXT_JA(context);
	cn->client_window = win;
}

void candwin_selection_changed(GtkWidget *menu_item, IMContextJA *cn) {
	int selected = 0;
	GList *candidate = cn->candidate_list;
	gchar *new_preedit;
	GtkWidget *label;
	gchar *string;
	
	//get the label of the menuitem
	label = gtk_container_get_children(GTK_CONTAINER(menu_item))->data;
	string = (gchar *)gtk_label_get_text(GTK_LABEL(label));
	//	printf("new preedit: %s\n", utf82euc(string));

	//FIXME: we need to store the candidate number, not the string because we can have doubles
	while (strcmp(candidate->data, string) != 0) {
		selected++;
		candidate = g_list_next(candidate);
	}
	if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_WNN)) {
#ifndef DISABLE_WNN
		im_ja_wnn_select_candidate(cn, selected);
		im_ja_wnn_update_preedit(cn);
#endif
	}
	else {
		new_preedit = g_new0(gchar, BUFFERSIZE);
		strncpy(new_preedit, cn->preedit_buf, cn->preedit_reverse_start);
		g_strlcat(new_preedit, string, BUFFERSIZE);
		g_strlcat(new_preedit, cn->preedit_buf + cn->preedit_reverse_end, BUFFERSIZE);
		cn->preedit_reverse_end = strlen(string);
		g_strlcpy(cn->preedit_buf, new_preedit, BUFFERSIZE);
		cn->preedit_reverse_end = cn->preedit_reverse_start + strlen(string);
		// printf("new preedit: %s\n", utf82euc(cn->preedit_buf));
		g_free(new_preedit);
		g_signal_emit_by_name(cn, "preedit_changed");
	}
	
}


void cand_selected(GtkWidget *menu_item, IMContextJA *cn) {
	int selected = 0;
	GList *candidate = cn->candidate_list;

	GtkWidget *label;
	gchar *string;

	//printf("selected. commit \n");
	
	//get the label of the menuitem
	label = gtk_container_get_children(GTK_CONTAINER(menu_item))->data;
	string = (gchar *)gtk_label_get_text(GTK_LABEL(label));
	//	printf("new preedit: %s\n", utf82euc(string));

	while (strcmp(candidate->data, string) != 0) {
		selected++;
		candidate = g_list_next(candidate);
	}
	//printf("selected: %d\n", selected);
	if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_CANNA)) {
#ifndef DISABLE_CANNA
		//im_ja_canna_commit(cn);
		IM_JA_DEBUG("Candidate window calling canna_rk_choose_candidate(%d)\n",selected);
		canna_rk_choose_candidate(cn,selected);
#endif
	}
	else 	if ((cn->input_method == IM_JA_KANJI_INPUT) && (cfg.default_conv_engine == KANJI_INPUT_WNN)) {
#ifndef DISABLE_WNN
		im_ja_wnn_select_candidate(cn, selected);
		im_ja_wnn_update_preedit(cn);
#endif
	}
	else {
		g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
	}
	im_ja_hide_candwin(cn);
	im_ja_show_status_window(cn);
}

void im_ja_free_candidate_list(IMContextJA *cn) {
	GList *tmplist;
	
	tmplist = cn->candidate_list;
	if (tmplist == NULL) return;

	while (tmplist != NULL) {
		g_free(tmplist->data);
		tmplist = g_list_next(tmplist);
	}
	g_list_free(cn->candidate_list);
	cn->candidate_list = NULL;
}

void im_ja_reset(IMContextJA *cn) {
	IM_JA_DEBUG("im_ja_reset()\n");
	*cn->preedit_insert = 0;

	if (cn->preedit_buf == NULL) cn->preedit_buf = g_new0(guchar, BUFFERSIZE);
	if (cn->preedit_insert == NULL) cn->preedit_insert = g_new0(guchar, BUFFERSIZE);
	if (cn->kakutei_buf == NULL) cn->kakutei_buf = g_new0(guchar, BUFFERSIZE);
	
	memset(cn->preedit_buf, 0, BUFFERSIZE);
	cn->preedit_reverse_start = 0;
	cn->preedit_reverse_end = 0;
	memset(cn->kakutei_buf, 0, BUFFERSIZE);
	cn->cand_stat = 0;
	cn->cursor_ndx = -1;
	cn->cursor_wnn = 0;
	g_signal_emit_by_name(cn, "preedit_changed");
	g_signal_emit_by_name(cn, "preedit_end");
	cn->candwin_pos_offset_x = cn->cursor_pos_offset_x; // Cursor position does not always change
	cn->candwin_pos_offset_y = cn->cursor_pos_offset_y; // so we need to update manually
}

void set_candwin_position(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, IMContextJA *cn) {
	gint tl_w = 0, tl_h = 0; //toplevel size
	gint tl_x = 0, tl_y = 0; //toplevel pos
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;
			
	if (cn->client_window != NULL) gdk_window_get_origin(cn->client_window, &cw_x, &cw_y);
	target_x = cw_x + cn->candwin_pos_offset_x;
	target_y = cw_y + cn->candwin_pos_offset_y;
	if (cn->toplevel != NULL) {
		/* check if target_ is within the toplevel window */
		//gtk_window_get_size(GTK_WINDOW(cn->toplevel), &tl_w, &tl_h);
		gdk_drawable_get_size(GDK_DRAWABLE(cn->toplevel), &tl_w, &tl_h);
		gdk_window_get_position(GDK_WINDOW(cn->toplevel), &tl_x, &tl_y);
		if (target_y > tl_y + tl_h) target_y = tl_y + tl_h;
		if (target_x > tl_x + tl_w) target_x = tl_x + tl_w;
	}

	/*
	*x = cw_x + cn->cursor_pos_offset_x;
	*y = cw_y + cn->cursor_pos_offset_y + 12;
	*/
	*x = target_x;
	*y = target_y + 12;

	IM_JA_DEBUG("candwin pos x: %d, y: %d\n", *x, *y);
}
static void internal_im_ja_hide_candwin(GtkWidget *candwin,IMContextJA *cn)
{
	IM_JA_DEBUG("internal_im_ja_hide_candwin (cn=0x%x)(cn->candwin=%x)\n",(unsigned)cn, (unsigned)candwin);
	im_ja_hide_candwin(cn);
}

void im_ja_show_candwin(IMContextJA *cn, int selected) {
	GtkWidget *candidate_item;
	GList *tmplist = NULL;
	gchar *tmpptr;

	IM_JA_DEBUG("FIXME: show_candwin() set selection\n");

	if (cn->candwin != NULL) {
		gtk_widget_destroy(cn->candwin);
	}
	cn->candwin = gtk_menu_new();
	g_signal_connect_after(GTK_OBJECT(cn->candwin), "destroy", G_CALLBACK(internal_im_ja_hide_candwin), (void*)cn);
	g_signal_connect_after(GTK_OBJECT(cn->candwin), "cancel", G_CALLBACK(internal_im_ja_hide_candwin), (void*)cn);
 
	tmplist = cn->candidate_list;
	while (tmplist != NULL) {
		tmpptr = (char *)tmplist->data;
		// printf("tmplist:%d, data:%d\n", tmplist, tmpptr);
		if (tmplist->data != NULL) {
			//printf("cand: %s\n", utf82euc(tmpptr));
			candidate_item = gtk_menu_item_new_with_label(tmplist->data);
			
			gtk_menu_shell_append(GTK_MENU_SHELL(cn->candwin), candidate_item);
			g_signal_connect(GTK_OBJECT(candidate_item), "select", 
											 G_CALLBACK(candwin_selection_changed), cn);
			g_signal_connect(GTK_OBJECT(candidate_item), "activate", 
											 G_CALLBACK(cand_selected), cn);
			tmplist = g_list_next(tmplist);
		}
		else {
			IM_JA_DEBUG("null received in GList\n");
			break;
		}
	}

	gtk_widget_show_all(cn->candwin);
	gtk_menu_popup(GTK_MENU(cn->candwin), NULL, NULL, (GtkMenuPositionFunc) set_candwin_position, cn, 0, 
								 gtk_get_current_event_time());

	im_ja_hide_status_window(cn);
	im_ja_hide_kanjipad(cn);
}

void im_ja_hide_candwin(IMContextJA* cn) {
	//IM_JA_DEBUG("hide candwin. (cn=0x%x)\n", (unsigned int)cn);
	cn->candwin = NULL;
	im_ja_show_status_window(cn);
}

void disable_status_window_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA* cn) {
	IM_JA_DEBUG("status window got focus.\n");
	//gtk_grab_add(cn->status_window);
	cn->status_window_focus_out_disabled = TRUE;
}
void disable_kanjipad_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA* cn ) {
	cn->kanjipad_focus_out_disabled = TRUE;
}

/*
	void keyevent_proxy(GtkWidget *widget) { //, GdkEventKey *event, IMContextJA* cn) {
	IM_JA_DEBUG("proxying keyevent.\n");
	//im_ja_filter_keypress((GtkIMContext *)cn, event); 
	}
*/

void im_ja_show_status_window(IMContextJA* cn) {
	//	IM_JA_DEBUG("show status win\n");
	
	if (cfg.status_win_position == STATUS_DISABLED) return;

	if (cn->show_first == FALSE) return; //don't show until we get focus once
	
	if (cn->candwin != NULL) {
		if (GTK_WIDGET_VISIBLE(cn->candwin) == TRUE)	return;
	}
	 
	if ((cn->input_method == IM_JA_RAW_INPUT) || (cn->input_method == IM_JA_KANJIPAD_INPUT))	{
		return;
	}

	if (cn->status_window != NULL) {
		//IM_JA_DEBUG("setting label to %s\n", cfg.status_win_labels[cn->input_method]);
		//		gtk_widget_set_default_size(cn->status_label, 40, 20);
		gtk_label_set_text(GTK_LABEL(cn->status_label), cfg.status_win_labels[cn->input_method]);
	}
	else {
		//IM_JA_DEBUG("creating status win\n");
		cn->status_window = gtk_window_new(GTK_WINDOW_POPUP);
		//g_signal_connect(G_OBJECT(cn->status_window), "key_release_event", G_CALLBACK(keyevent_proxy), (gpointer) cn);
		gtk_window_set_title(GTK_WINDOW(cn->status_window), "im");
		gtk_window_set_default_size(GTK_WINDOW(cn->status_window), 40, 25);
		cn->status_border = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(cn->status_border), GTK_SHADOW_OUT);
		gtk_container_add(GTK_CONTAINER(cn->status_window), cn->status_border);
		//IM_JA_DEBUG("setting label to %s\n", cfg.status_win_labels[cn->input_method]);
		//		gtk_widget_set_default_size(cn->status_label, 40, 20);
		cn->status_label = gtk_label_new(cfg.status_win_labels[cn->input_method]);
		gtk_container_add(GTK_CONTAINER(cn->status_border), cn->status_label);
		gtk_widget_add_events(cn->status_label, GDK_ALL_EVENTS_MASK);

		/* This hack is to disable flashy behaviour when status window gets the focus*/
		g_signal_connect(GTK_OBJECT(cn->status_window), "enter_notify_event", 
										 GTK_SIGNAL_FUNC(disable_status_window_focus_out), cn);

		cn->toplevel = gdk_window_get_toplevel(cn->client_window);

		im_ja_update_status_window_location(cn);
	}
	gtk_widget_show_all(cn->status_window);
}

void im_ja_hide_status_window(IMContextJA *cn) {
	if (cn->status_window != NULL) {
		gtk_widget_hide_all(cn->status_window);
	}
}

void im_ja_show_kanjipad(IMContextJA *cn) {
	if (cn->show_first == FALSE) return; //don't show until we get focus once
	im_ja_reset(cn);

	if (cn->kanjipad == NULL) {
		int w = 80, h = 70;
		if (cfg.kanjipad_custom_size == TRUE) {
			w = cfg.kanjipad_size[0];
			h = cfg.kanjipad_size[1];
		}
		if (create_kanjipad_win(cn, w, h) != NULL) {
			/* This hack is to disable flashy behaviour when the window gets the focus*/
			g_signal_connect(GTK_OBJECT(cn->kanjipad), "enter_notify_event", 
											 GTK_SIGNAL_FUNC(disable_kanjipad_focus_out), cn);

			im_ja_set_kanjipad_location(cn);
		}
		else {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		}
	}
	
	if (cn->kanjipad != NULL) {
		gtk_widget_show_all(cn->kanjipad);
	}
}

void im_ja_hide_kanjipad(IMContextJA *cn) {
	if (cn->kanjipad != NULL) gtk_widget_hide_all(cn->kanjipad);
}


static void im_ja_cursor_location_changed(GtkIMContext *context, GdkRectangle *area) {
	IMContextJA* cn = IM_CONTEXT_JA(context);
	gint x, y, area_h;

	//Bug in GTK?:	width & height is sometimes (on first set_location) a huge_number
	area_h = area->height;
	if ((area->width > 1000) || (area->width < 0) ||
			(area->height > 1000) || (area->height < 0)) {
		area_h = 12;
	}
	x = area->x;
	y = area->y + area_h/2;
	
	//IM_JA_DEBUG("cursor location: %d, %d\n", x, y);
	if ((cn->cursor_pos_offset_x != x) || (cn->cursor_pos_offset_y != y)) {
		if ((strlen(cn->preedit_buf) == 0) || (cn->update_candwin_pos == TRUE)) {
			cn->candwin_pos_offset_x = x;
			cn->candwin_pos_offset_y = y;
			cn->update_candwin_pos = FALSE;
			//IM_JA_DEBUG("candwin cursor_location_changed(%d, %d)\n", x, y);
		}
		cn->cursor_pos_offset_x = x;
		cn->cursor_pos_offset_y = y;
		//IM_JA_DEBUG("cursor_location_changed(%d, %d)\n", x, y);
		im_ja_set_kanjipad_location(cn);
		im_ja_update_status_window_location(cn);
	}
}

void im_ja_update_status_window_location(IMContextJA *cn) {
	gint status_w, status_h;
	gint tl_w = 0, tl_h = 0; //toplevel size
	gint tl_x = 0, tl_y = 0; //toplevel pos
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;
			
	if (cn->status_window !=	NULL) {
		if (cfg.status_win_position == STATUS_DISABLED) return;
		
		if (cfg.status_win_position == STATUS_BELOW_CURSOR) {
			if (cn->client_window != NULL) gdk_window_get_origin(cn->client_window, &cw_x, &cw_y);
			target_x = cw_x + cn->cursor_pos_offset_x;
			target_y = cw_y + cn->cursor_pos_offset_y;
			if (cn->toplevel != NULL) {
				/* check if target_ is within the toplevel window */
				gdk_drawable_get_size(GDK_DRAWABLE(cn->toplevel), &tl_w, &tl_h);
				gdk_window_get_position(GDK_WINDOW(cn->toplevel), &tl_x, &tl_y);
				if ((tl_w != 0) || (tl_h != 0)) {
					if (target_y > tl_y + tl_h) target_y = tl_y + tl_h - 12;
					if (target_x > tl_x + tl_w) target_x = tl_x + tl_w;
				}
				else IM_JA_DEBUG("Couldn't get toplevel window dimensions!\n"); 
			}
			IM_JA_DEBUG("toplevel window dimensions. Got width: %d, height: %d, x: %d, y: %d\n", tl_w, tl_h, tl_x, tl_y); 
			IM_JA_DEBUG("update status window location x: %d, y: %d\n", target_x, target_y);
			gtk_window_get_size(GTK_WINDOW(cn->status_window), &status_w, &status_h);
			gtk_window_move((GTK_WINDOW(cn->status_window)), target_x - status_w/2, target_y + 12);
			return;
		}
		
		if (cfg.status_win_position == STATUS_WINDOW_BOTTOM_LEFT) {
			if (cn->toplevel != NULL) {
				/* The following is for attaching the status window to the window border if toplevel is a gtkwindow
				//				GdkGravity gravity = gtk_window_get_gravity(GTK_WINDOW(cn->toplevel));
				//				gtk_window_set_gravity(GTK_WINDOW(cn->toplevel), GDK_GRAVITY_SOUTH_WEST);
				//				gtk_window_set_gravity(GTK_WINDOW(cn->toplevel), gravity);
				//				gtk_window_get_size(GTK_WINDOW(cn->toplevel), &tl_w, &tl_h);
				*/
				gdk_window_get_position(GDK_WINDOW(cn->toplevel), &tl_x, &tl_y);
				gdk_drawable_get_size(GDK_DRAWABLE(cn->toplevel), &tl_w, &tl_h);
		
				if ((tl_w != 0) || (tl_h != 0)) {
					gtk_window_move((GTK_WINDOW(cn->status_window)), tl_x, tl_y + tl_h); 
				}
				else { /* use the client window instead */
					gint cw_h = 0, cw_w = 0;
					IM_JA_DEBUG("Couldn't get toplevel window dimensions!\n"); 
					if (cn->client_window != NULL) gdk_window_get_origin(cn->client_window, &cw_x, &cw_y);
					gdk_drawable_get_size(GDK_DRAWABLE(cn->client_window), &cw_w, &cw_h);
					gtk_window_move((GTK_WINDOW(cn->status_window)), cw_x, cw_y + cw_h); 
				}
				IM_JA_DEBUG("toplevel window dimensions. Got width: %d, height: %d, x: %d, y: %d\n", tl_w, tl_h, tl_x, tl_y); 
				//IM_JA_DEBUG("toplevel window title: %s\n", gtk_window_get_title(GTK_WINDOW(cn->toplevel)));
			}
			return;
		}
		
		if (cfg.status_win_position == STATUS_FIXED_COORD) {
			gtk_window_move((GTK_WINDOW(cn->status_window)), cfg.status_win_coords[0], cfg.status_win_coords[1]);
			return;
		}
	}
}

void im_ja_set_kanjipad_location(IMContextJA *cn) {
	gint _w, _h;
	gint tl_w = 0, tl_h = 0; //toplevel size
	gint tl_x = 0, tl_y = 0; //toplevel pos
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;
			
	if (cn->kanjipad !=	 NULL) {
		if (cn->client_window != NULL) gdk_window_get_origin(cn->client_window, &cw_x, &cw_y);
		target_x = cw_x + cn->cursor_pos_offset_x;
		target_y = cw_y + cn->cursor_pos_offset_y;
		if (cn->toplevel != NULL) {
			/* check if target_ is within the toplevel window */
			gdk_drawable_get_size(GDK_DRAWABLE(cn->toplevel), &tl_w, &tl_h);
			gdk_window_get_position(GDK_WINDOW(cn->toplevel), &tl_x, &tl_y);

			if ((tl_w != 0) || (tl_h != 0)) {
				if (target_y > tl_y + tl_h) target_y = tl_y + tl_h;
				if (target_x > tl_x + tl_w) target_x = tl_x + tl_w;
			}
		}
		gtk_window_get_size(GTK_WINDOW(cn->kanjipad), &_w, &_h);
		gtk_window_move((GTK_WINDOW(cn->kanjipad)), target_x - _w/3, target_y + 12);
	}
}


static void im_ja_lost_focus(GtkIMContext *context) {
	IMContextJA* cn = IM_CONTEXT_JA(context);
	//IM_JA_DEBUG("im-ja lost focus\n");
	if (cn->status_window_focus_out_disabled != TRUE) {
		im_ja_hide_status_window(cn);
	}
	if (cn->kanjipad_focus_out_disabled != TRUE) {
		im_ja_hide_kanjipad(cn);
	}
}

static void im_ja_got_focus(GtkIMContext *context) {
	gint cw_x, cw_y;
	IMContextJA* cn = IM_CONTEXT_JA(context);

	//IM_JA_DEBUG("im-ja got focus\n");
	//	if (cn->status_window != NULL) gtk_grab_remove(cn->status_window);
	cn->status_window_focus_out_disabled = FALSE;
	if (cn->show_first == FALSE) {
		cn->show_first = TRUE;
	}

	if (cn->client_window != NULL) gdk_window_get_origin(cn->client_window, &cw_x, &cw_y);
	im_ja_update_status_window_location(cn);
	im_ja_show_status_window(cn);
	im_ja_set_kanjipad_location(cn);
	cn->kanjipad_focus_out_disabled = FALSE;
	if (cn->input_method == IM_JA_KANJIPAD_INPUT) {
		im_ja_show_kanjipad(cn);
	}
}

static void im_ja_on_reset(IMContextJA *context) {
	IM_JA_DEBUG("got \"reset\" event. im_ja_on_reset()\n");

	//Cancel preedit buffer
	if (strlen(context->preedit_buf) > 0) {
		g_signal_emit_by_name(context, "commit", context->preedit_buf);
		im_ja_reset(context);
		if (context->on_reset) {
			(*context->on_reset)(context);
		}
	}
}

gboolean im_ja_set_kanji_input(IMContextJA *cn, gint conv_engine) {
	IM_JA_DEBUG("im_ja_set_kanji_input\n");

	if (conv_engine == KANJI_INPUT_WNN) {
#ifndef DISABLE_WNN
		cn->on_reset = im_ja_wnn_reset;
		if (im_ja_wnn_init(cn) != TRUE) {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		}
#else
		im_ja_print_error(_("Wnn support is disabled. If you want wnn support, recompile im-ja."));
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return FALSE;
#endif
	}
	else if (conv_engine == KANJI_INPUT_CANNA) {
#ifndef DISABLE_CANNA
		cn->on_reset = canna_rk_reset;
		if (canna_rk_init(cn) != TRUE) {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		}
#else
		im_ja_print_error(_("Canna support is disabled. If you want canna support, recompile im-ja."));
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return FALSE;
#endif
	}
	cn->input_method = IM_JA_KANJI_INPUT;
	im_ja_show_status_window(cn);
	return TRUE;
}

gboolean im_ja_set_input_method(IMContextJA *cn, int input_method) {

	//IM_JA_DEBUG("setting input method: %d\n", input_method);

	/* Commit the preedit buffer */
	if (input_method != cn->input_method) {
		if (strlen(cn->preedit_buf) > 0) g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
	}

	cn->on_reset = NULL;

	cn->input_method = input_method;
	im_ja_hide_candwin(cn);
	im_ja_reset(cn);

	if (cn->input_method != IM_JA_KANJIPAD_INPUT) im_ja_hide_kanjipad(cn);
	switch (cn->input_method) {
	case IM_JA_KANJI_INPUT:
		return im_ja_set_kanji_input(cn, cfg.default_conv_engine);
		break;
	case IM_JA_KANJIPAD_INPUT:
		if (cfg.kanjipad_enabled == TRUE) {
			im_ja_hide_status_window(cn);
			im_ja_show_kanjipad(cn);
		}
		else {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		}
		break;
	case IM_JA_RAW_INPUT:
		im_ja_hide_status_window(cn);
		break;
	default: 
		im_ja_show_status_window(cn);
	}
	return TRUE;
}

void im_ja_next_input_method(IMContextJA *cn) {
	gint tmp;
	tmp = cn->input_method;
	tmp++;
	if ((tmp == IM_JA_KANJIPAD_INPUT) && (cfg.kanjipad_enabled == FALSE)) tmp++;
	if (tmp >= IM_JA_INPUT_METHODS_TOTAL) tmp = IM_JA_RAW_INPUT;
	im_ja_set_input_method(cn, tmp);
}

void im_ja_prev_input_method(IMContextJA *cn) {
	gint tmp;
	tmp = cn->input_method;
	tmp--;
	if ((tmp == IM_JA_KANJIPAD_INPUT) && (cfg.kanjipad_enabled == FALSE)) tmp--;
	if (tmp < IM_JA_RAW_INPUT) tmp = IM_JA_INPUT_METHODS_TOTAL - 1;
	im_ja_set_input_method(cn, tmp);
}

 
